//-------------------------------------------------------------------------------------------------
// engineAssert.h (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
// 
// Assertion macros.
//-------------------------------------------------------------------------------------------------
#ifndef _include_engineAssert_
#define _include_engineAssert_


//-------------------------------------------------------------------------------------------------
// Included header files
//-------------------------------------------------------------------------------------------------
#include "engineCommon.h"


//-------------------------------------------------------------------------------------------------
// String constants
//-------------------------------------------------------------------------------------------------
#define __DEBUG_ONLY__ "DEBUG ONLY"
#define __ALL_BUILDS__ "ALL BUILDS"


//-------------------------------------------------------------------------------------------------
// Function prototypes
//-------------------------------------------------------------------------------------------------
__declspec( noreturn ) void AssertionFailure( const int line, const char* const fileName, const char* const assertType, const char* const expression, const char* const format, ... );


//-------------------------------------------------------------------------------------------------
// ASSERT
//-------------------------------------------------------------------------------------------------
#if defined( ENABLE_ASSERTIONS )
#define ASSERT( exp )											\
	{															\
		if( !(exp) )											\
		AssertionFailure( __LINE__, __FILE__, __DEBUG_ONLY__,	\
			#exp, "" );											\
	}
#else
#define ASSERT( exp )
#endif


//-------------------------------------------------------------------------------------------------
// ASSERT_MSG0
//-------------------------------------------------------------------------------------------------
#if defined( ENABLE_ASSERTIONS )
#define ASSERT_MSG0( exp, msg )									\
	{															\
		if( !(exp) )											\
		AssertionFailure( __LINE__, __FILE__, __DEBUG_ONLY__,	\
			#exp, msg );										\
	}
#else
#define ASSERT_MSG0( exp, msg )
#endif


//-------------------------------------------------------------------------------------------------
// ASSERT_MSG1
//-------------------------------------------------------------------------------------------------
#if defined( ENABLE_ASSERTIONS )
#define ASSERT_MSG1( exp, msg, arg1 )							\
	{															\
		if( !(exp) )											\
		AssertionFailure( __LINE__, __FILE__, __DEBUG_ONLY__,	\
			#exp, msg, arg1 );									\
	}
#else
#define ASSERT_MSG1( exp, msg, arg1 )
#endif


//-------------------------------------------------------------------------------------------------
// ASSERT_MSG2
//-------------------------------------------------------------------------------------------------
#if defined( ENABLE_ASSERTIONS )
#define ASSERT_MSG2( exp, msg, arg1, arg2 )						\
	{															\
		if( !(exp) )											\
		AssertionFailure( __LINE__, __FILE__, __DEBUG_ONLY__,	\
			#exp, msg, arg1, arg2 );							\
	}
#else
#define ASSERT_MSG2( exp, msg, arg1, arg2 )
#endif


//-------------------------------------------------------------------------------------------------
// ASSERT_MSG3
//-------------------------------------------------------------------------------------------------
#if defined( ENABLE_ASSERTIONS )
#define ASSERT_MSG3( exp, msg, arg1, arg2, arg3 )				\
	{															\
		if( !(exp) )											\
		AssertionFailure( __LINE__, __FILE__, __DEBUG_ONLY__,	\
			#exp, msg, arg1, arg2, arg3 );						\
	}
#else
#define ASSERT_MSG3( exp, msg, arg1, arg2, arg3 )
#endif


//-------------------------------------------------------------------------------------------------
// ASSERT_MSG4
//-------------------------------------------------------------------------------------------------
#if defined( ENABLE_ASSERTIONS )
#define ASSERT_MSG4( exp, msg, arg1, arg2, arg3, arg4 )			\
	{															\
		if( !(exp) )											\
		AssertionFailure( __LINE__, __FILE__, __DEBUG_ONLY__,	\
			#exp, msg, arg1, arg2, arg3, arg4 );				\
	}
#else
#define ASSERT_MSG4( exp, msg, arg1, arg2, arg3, arg4 )
#endif


//-------------------------------------------------------------------------------------------------
// ASSERT_MSG5
//-------------------------------------------------------------------------------------------------
#if defined( ENABLE_ASSERTIONS )
#define ASSERT_MSG5( exp, msg, arg1, arg2, arg3, arg4, arg5 )	\
	{															\
		if( !(exp) )											\
		AssertionFailure( __LINE__, __FILE__, __DEBUG_ONLY__,	\
			#exp, msg, arg1, arg2, arg3, arg4, arg5 );			\
	}
#else
#define ASSERT_MSG5( exp, msg, arg1, arg2, arg3, arg4, arg5 )
#endif


//-------------------------------------------------------------------------------------------------
// VALIDATE
//-------------------------------------------------------------------------------------------------
#define VALIDATE( exp )											\
	{															\
		if( !(exp) )											\
		AssertionFailure( __LINE__, __FILE__, __ALL_BUILDS__,	\
			#exp, "" );											\
	}


//-------------------------------------------------------------------------------------------------
// VALIDATE_MSG0
//-------------------------------------------------------------------------------------------------
#define VALIDATE_MSG0( exp, msg )								\
	{															\
		if( !(exp) )											\
		AssertionFailure( __LINE__, __FILE__, __ALL_BUILDS__,	\
			#exp, msg );										\
	}


//-------------------------------------------------------------------------------------------------
// VALIDATE_MSG1
//-------------------------------------------------------------------------------------------------
#define VALIDATE_MSG1( exp, msg, arg1 )							\
	{															\
		if( !(exp) )											\
		AssertionFailure( __LINE__, __FILE__, __ALL_BUILDS__,	\
			#exp, msg, arg1 );									\
	}


//-------------------------------------------------------------------------------------------------
// VALIDATE_MSG2
//-------------------------------------------------------------------------------------------------
#define VALIDATE_MSG2( exp, msg, arg1, arg2 )					\
	{															\
		if( !(exp) )											\
		AssertionFailure( __LINE__, __FILE__, __ALL_BUILDS__,	\
			#exp, msg, arg1, arg2 );							\
	}


//-------------------------------------------------------------------------------------------------
// VALIDATE_MSG3
//-------------------------------------------------------------------------------------------------
#define VALIDATE_MSG3( exp, msg, arg1, arg2, arg3 )				\
	{															\
		if( !(exp) )											\
		AssertionFailure( __LINE__, __FILE__, __ALL_BUILDS__,	\
			#exp, msg, arg1, arg2, arg3 );						\
	}


//-------------------------------------------------------------------------------------------------
// VALIDATE_MSG4
//-------------------------------------------------------------------------------------------------
#define VALIDATE_MSG4( exp, msg, arg1, arg2, arg3, arg4 )		\
	{															\
		if( !(exp) )											\
		AssertionFailure( __LINE__, __FILE__, __ALL_BUILDS__,	\
			#exp, msg, arg1, arg2, arg3, arg4 );				\
	}


//-------------------------------------------------------------------------------------------------
// VALIDATE_MSG5
//-------------------------------------------------------------------------------------------------
#define VALIDATE_MSG5( exp, msg, arg1, arg2, arg3, arg4, arg5 )	\
	{															\
		if( !(exp) )											\
		AssertionFailure( __LINE__, __FILE__, __ALL_BUILDS__,	\
			#exp, msg, arg1, arg2, arg3, arg4, arg5 );			\
	}


//-------------------------------------------------------------------------------------------------
// PANIC
//-------------------------------------------------------------------------------------------------
#define PANIC( msg )											\
	{															\
		AssertionFailure( __LINE__, __FILE__, __ALL_BUILDS__,	\
			"", msg );											\
	}


//-------------------------------------------------------------------------------------------------
// PANIC_MSG0
//-------------------------------------------------------------------------------------------------
#define PANIC_MSG0( msg )										\
	{															\
		AssertionFailure( __LINE__, __FILE__, __ALL_BUILDS__,	\
			"", msg );											\
	}


//-------------------------------------------------------------------------------------------------
// PANIC_MSG1
//-------------------------------------------------------------------------------------------------
#define PANIC_MSG1( msg, arg1 )									\
	{															\
		AssertionFailure( __LINE__, __FILE__, __ALL_BUILDS__,	\
			"", msg, arg1 );									\
	}


//-------------------------------------------------------------------------------------------------
// PANIC_MSG2
//-------------------------------------------------------------------------------------------------
#define PANIC_MSG2( msg, arg1, arg2 )							\
	{															\
		AssertionFailure( __LINE__, __FILE__, __ALL_BUILDS__,	\
			"", msg, arg1, arg2 );								\
	}


//-------------------------------------------------------------------------------------------------
// PANIC_MSG3
//-------------------------------------------------------------------------------------------------
#define PANIC_MSG3( msg, arg1, arg2, arg3 )						\
	{															\
		AssertionFailure( __LINE__, __FILE__, __ALL_BUILDS__,	\
			"", msg, arg1, arg2, arg3 );						\
	}


//-------------------------------------------------------------------------------------------------
// PANIC_MSG4
//-------------------------------------------------------------------------------------------------
#define PANIC_MSG4( msg, arg1, arg2, arg3, arg4 )				\
	{															\
		AssertionFailure( __LINE__, __FILE__, __ALL_BUILDS__,	\
			"", msg, arg1, arg2, arg3, arg4 );					\
	}


//-------------------------------------------------------------------------------------------------
// PANIC_MSG5
//-------------------------------------------------------------------------------------------------
#define PANIC_MSG5( msg, arg1, arg2, arg3, arg4, arg5 )			\
	{															\
		AssertionFailure( __LINE__, __FILE__, __ALL_BUILDS__,	\
			"", msg, arg1, arg2, arg3, arg4, arg5 );			\
	}


#endif // _include_engineAssert_
