// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.

#include "pch.h"
#include "gameCommon.h"
#include "EnemyGuard.h"
#include "GameRenderer.h"
#include "TheGame.h"
#include "Player.h"
#include "PathFinding.h"


std::list< EnemyGuard* > EnemyGuard::m_guardList;




//-------------------------------------------------------------------------------------------------
EnemyGuard::EnemyGuard()
: m_origin( 0, 0 ),
  m_forwardDeg( 0 ),
  m_visionDeg( 0 ),
  m_fov( ENEMY_GUARD_DEFAULT_FOV ),
  m_visionDist( ENEMY_GUARD_DEFAULT_VISIONDIST ),
  m_goalGridX( 0 ),
  m_goalGridY( 0 )
{
}

//-------------------------------------------------------------------------------------------------
EnemyGuard::EnemyGuard( const Vector2& origin, const float forwardDeg )
: m_origin( origin ),
  m_forwardDeg( forwardDeg ),
  m_visionDeg( forwardDeg ),
  m_fov( ENEMY_GUARD_DEFAULT_FOV ),
  m_visionDist( ENEMY_GUARD_DEFAULT_VISIONDIST ),
  m_goalGridX( 0 ),
  m_goalGridY( 0 )
{
	m_forward.SetFromCosSinDegrees( forwardDeg );

	/// Find the ACTUAL starting position of this enemy; that is, on the nearest path
	int startX, startY;
	g_pathFinding.FindClosestPathLocation( startX, startY, origin );
	m_origin.SetXY( (float) startX, (float) startY );
	m_goalGridX = startX;
	m_goalGridY = startY;
}


//-------------------------------------------------------------------------------------------------
EnemyGuard::~EnemyGuard()
{
}

//-------------------------------------------------------------------------------------------------
void EnemyGuard::Render( const bool highlight )
{
	const float yborder = (float) (ENEMY_GUARD_HEIGHT-1) / 2;
	const float xborder = (float) (ENEMY_GUARD_WIDTH-1) / 2;
	const float top = m_origin.y - yborder;
	const float left = m_origin.x - xborder;
	const float right = m_origin.x + xborder;
	const float bottom = m_origin.y + yborder;
	glBegin( GL_QUADS );
	glColor3f( 1.0f, 0.0f, 0.0f );
	glVertex2f( left, top );
	glVertex2f( left , bottom );
	glVertex2f( right, bottom );
	glVertex2f( right, top );
	glEnd();

	g_renderer.RenderVisionCone( m_origin, m_visionDeg, m_visionDist, m_fov, m_canSeePlayer );
//	g_renderer.RenderVisionCone( m_origin, m_forwardDeg, m_visionDist, m_fov, m_canSeePlayer );
}

//-------------------------------------------------------------------------------------------------
void EnemyGuard::MoveForward( const float dist )
{
	float clampedDist = Min( dist, 1.0f );
	m_origin += m_forward * clampedDist;
}

//-------------------------------------------------------------------------------------------------
void EnemyGuard::Turn( const float deg )
{
	m_forwardDeg += deg;
	while( m_forwardDeg < 0 )	{ m_forwardDeg += 360; }
	while( m_forwardDeg > 359 ) { m_forwardDeg -= 360; }
	m_forward.SetFromCosSinDegrees( m_forwardDeg );
}


//-------------------------------------------------------------------------------------------------
void EnemyGuard::LerpVisionDeg()
{
	/// TEMP - no interpolation until it works well
//	m_visionDeg = m_forwardDeg;
//	return;



	float turnDeltaDegrees = 0.0f;
	const float ENEMY_GUARD_TURN_RATE_DEGREES_PER_SECOND = 600.0f;
	float turnDegrees = ENEMY_GUARD_TURN_RATE_DEGREES_PER_SECOND * (float)APPLICATION->GetFrameSeconds();

	float angularDistanceDegrees = (float) CalcAngularDistanceDegrees( m_visionDeg, m_forwardDeg );
//	float angularDistanceDegrees = m_forwardDeg - m_visionDeg;

/*
	const float THRESHOLD_ANGLE = 0.50f;
	const float MINIMUM_MULTIPLIER = 0.1f;
	float absAngularDistanceDegrees = (float) fabs( angularDistanceDegrees );
	if( absAngularDistanceDegrees < THRESHOLD_ANGLE )
	{
		turnDegrees *= MINIMUM_MULTIPLIER;
	}
	else
	{
		float overThresholdFraction = (absAngularDistanceDegrees - THRESHOLD_ANGLE) / (180.0f - THRESHOLD_ANGLE);
		float multiplier = MINIMUM_MULTIPLIER + ((1.0f - MINIMUM_MULTIPLIER) * overThresholdFraction);
		turnDegrees *= multiplier;
	}
/*/

	float multiplier = 1.0f;
	float absAngularDistanceDegrees = (float) fabs( angularDistanceDegrees );
	if( absAngularDistanceDegrees > 95.0f )
	{
		multiplier = 1.0f;
	}
	else if( absAngularDistanceDegrees > 50.0f )
	{
		multiplier = 0.5f;
	}
	else if( absAngularDistanceDegrees > 35.0f )
	{
		multiplier = 0.2f;
	}
	else if( absAngularDistanceDegrees > 15.0f )
	{
		multiplier = 0.1f;
	}
	else
	{
		multiplier = 0.05f;
	}

	turnDegrees *= multiplier;
/**/

	if( angularDistanceDegrees > turnDegrees )
	{
		turnDeltaDegrees = turnDegrees;
	}
	else if( angularDistanceDegrees > 0.0f )
	{
		turnDeltaDegrees = angularDistanceDegrees;
	}
	else if( angularDistanceDegrees < -turnDegrees )
	{
		turnDeltaDegrees = -turnDegrees;
	}
	else if( angularDistanceDegrees < 0.0f )
	{
		turnDeltaDegrees = angularDistanceDegrees;
	}

	m_visionDeg += turnDeltaDegrees;
	while( m_visionDeg > 360.0f )
	{
		m_visionDeg -= 360.0f;
	}
	while( m_visionDeg < 0.0f )
	{
		m_visionDeg += 360.0f;
	}
}


//-------------------------------------------------------------------------------------------------
void EnemyGuard::Think()
{
//	Turn( 50 * (float)APPLICATION->GetFrameSeconds() );
//	MoveForward( 2 * (float)APPLICATION->GetFrameSeconds() );

	PathMove();

	/// Turn view more towards actual movement direction
	LerpVisionDeg();

	const bool incone = TheGame::IsPointInViewCone( m_origin, g_player->m_origin, m_visionDeg, m_fov, m_visionDist );
//	const bool incone = TheGame::IsPointInViewCone( m_origin, g_player->m_origin, m_forwardDeg, m_fov, m_visionDist );
	m_canSeePlayer = incone && g_mapData->TraceLine( 1, m_origin, g_player->m_origin );
}


//-------------------------------------------------------------------------------------------------
void EnemyGuard::PathMove()
{
	int posX = (int) m_origin.x;
	int posY = (int) m_origin.y;

	/// Make sure we're on the path now
//	VALIDATE_MSG2( g_pathFinding.IsXYOnEnemyPath( posX, posY ), "Enemy at (%d,%d) was not on the path!\n", posX, posY );

	if( !g_pathFinding.IsXYOnEnemyPath( posX, posY ) )
	{
		RENDERER->XYPrintf( 100, 100, "ENEMY OFF THE GRID at %d,%d !", posX, posY );
	}

	Vector2 vectorToGoal( m_forward );
	FindClosestPathXYForCurrentDirection2( posX, posY, m_goalGridX, m_goalGridY, vectorToGoal );
	if( vectorToGoal != Vector2::ZERO )
	{
		m_forward = vectorToGoal;
		m_forwardDeg = (float) m_forward.GetDegrees();
	}
	MoveForward( 20 * (float)APPLICATION->GetFrameSeconds() );
}


//-------------------------------------------------------------------------------------------------
void EnemyGuard::FindClosestPathXYForCurrentDirection( int posX, int posY, int& closestX, int& closestY, Vector2& vectorToGoal )
{
	/// Find the best adjacent pixel in the direction we're currently moving
	int bestDirection = -1;
	float bestDotProduct = -2.0f;
	int bestX = 0;
	int bestY = 0;

	VALIDATE_MSG2( posX > MAP_MIN_X, "Enemy hit the left edge of the map at (%d,%d)\n", posX, posY );
	VALIDATE_MSG2( posY > MAP_MIN_Y, "Enemy hit the top edge of the map at (%d,%d)\n", posX, posY );
	VALIDATE_MSG2( posX < MAP_MAX_X, "Enemy hit the right edge of the map at (%d,%d)\n", posX, posY );
	VALIDATE_MSG2( posY < MAP_MAX_Y, "Enemy hit the bottom edge of the map at (%d,%d)\n", posX, posY );

	for( int direction = 0; direction < 8; direction ++ )
	{
		int scanX = posX + g_semiCardinalX[ direction ];
		int scanY = posY + g_semiCardinalY[ direction ];
		if( g_mapData->data[ scanX ][ scanY ].type != GRIDTYPE_ENEMYPATH )
			continue;

		Vector2 ofsVector( g_semiCardinalCos[ direction ], g_semiCardinalSin[ direction ] );
		float dotProduct = ofsVector | m_forward;
		if( dotProduct > bestDotProduct )
		{
			bestDotProduct = dotProduct;
			bestDirection = direction;
			bestX = scanX;
			bestY = scanY;
		}
	}

	/// Make sure we found one
	if( bestDirection == -1 )
	{
		closestX = posX;
		closestY = posY;
		vectorToGoal = Vector2::ZERO;
		return;
	}

	vectorToGoal.SetXY( g_semiCardinalCos[ bestDirection ], g_semiCardinalSin[ bestDirection ] );
	closestX = bestX;
	closestY = bestY;
}



//-------------------------------------------------------------------------------------------------
void EnemyGuard::FindClosestPathXYForCurrentDirection2( int posX, int posY, int& closestX, int& closestY, Vector2& vectorToGoal )
{
	/// Find the best adjacent pixel in the direction we're currently moving
	int bestDirection = -1;
	float bestDotProduct = -2.0f;
	int bestX = 0;
	int bestY = 0;

	VALIDATE_MSG2( posX > MAP_MIN_X, "Enemy hit the left edge of the map at (%d,%d)\n", posX, posY );
	VALIDATE_MSG2( posY > MAP_MIN_Y, "Enemy hit the top edge of the map at (%d,%d)\n", posX, posY );
	VALIDATE_MSG2( posX < MAP_MAX_X, "Enemy hit the right edge of the map at (%d,%d)\n", posX, posY );
	VALIDATE_MSG2( posY < MAP_MAX_Y, "Enemy hit the bottom edge of the map at (%d,%d)\n", posX, posY );

	for( int direction = 0; direction < 16; direction ++ )
	{
		int scanX = posX + g_quarterCardinalX[ direction ];
		int scanY = posY + g_quarterCardinalY[ direction ];
		if( g_mapData->data[ scanX ][ scanY ].type != GRIDTYPE_ENEMYPATH )
			continue;

		Vector2 ofsVector( g_quarterCardinalCos[ direction ], g_quarterCardinalSin[ direction ] );
		float dotProduct = ofsVector | m_forward;
		if( dotProduct > bestDotProduct )
		{
			bestDotProduct = dotProduct;
			bestDirection = direction;
			bestX = scanX;
			bestY = scanY;
		}
	}

	/// Make sure we found one
	if( bestDirection == -1 )
	{
		closestX = posX;
		closestY = posY;
		vectorToGoal = Vector2::ZERO;
		return;
	}

	vectorToGoal.SetXY( g_quarterCardinalCos[ bestDirection ], g_quarterCardinalSin[ bestDirection ] );
	closestX = bestX;
	closestY = bestY;
}



