// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.

#include "pch.h"
#include "TheGame.h"
#include "MapData.h"


//-------------------------------------------------------------------------------------------------
// Global shit
//-------------------------------------------------------------------------------------------------
MapData* g_mapData = NULL;


GridDataAssociation gridDataTable[ NUM_GRIDTYPES ] =
{
//	 GRIDTYPE enum				COLOR enum			isVisible	description
	{GRIDTYPE_EMPTY,			COLOR_BLACK,		true,		"Empty"},
	{GRIDTYPE_WALL,				COLOR_WHITE,		true,		"Wall"},
	{GRIDTYPE_ENEMYSTART,		COLOR_RED,			false,		"EnemyStart"},
	{GRIDTYPE_ENEMYFACINGHINT,	COLOR_HALFRED,		false,		"EnemyFacingHint"},
	{GRIDTYPE_ENEMYPATH,		COLOR_BLUE,			false,		"EnemyPath"},
	{GRIDTYPE_PLAYERSTART,		COLOR_GREEN,		false,		"PlayerStart"},
	{GRIDTYPE_PLAYERGOAL,		COLOR_YELLOW,		false,		"PlayerGoal"},
};



//-------------------------------------------------------------------------------------------------
// GetMapDataTypeForColor
//-------------------------------------------------------------------------------------------------
GridDataType MapData::GetMapDataTypeForColor( Color& color )
{
	for( int i = 0; i < NUM_GRIDTYPES; i ++ )
	{
		if( gridDataTable[ i ].color == color )
		{
			return gridDataTable[ i ].type;
		}
	}

	return GRIDTYPE_UNKNOWN;
}


//-------------------------------------------------------------------------------------------------
// ParseImageMetaData
// 
// - Scans all pixels in the raw image data
//-------------------------------------------------------------------------------------------------
MapData::ParseImageMetaData( unsigned char* pixels, int width, int height )
{
	VALIDATE( width == MAP_WIDTH );
	VALIDATE( height == MAP_HEIGHT );

	for( int x = 0; x < width; x ++ )
	for( int y = 0; y < height; y ++ )
	{
		/// Extract the color
		Color color;
		int pixelNumber = (y * width) + x;
		int startByteNumber = pixelNumber * 3;
		color.b = pixels[ startByteNumber + 0 ];
		color.g = pixels[ startByteNumber + 1 ];
		color.r = pixels[ startByteNumber + 2 ];

		//VALIDATE_MSG3( 0, "TEST: first pixel (0,0) is colored %d,%d,%d\n", color.r, color.g, color.b );

		/// Convert color to grid type
		GridDataType type = GetMapDataTypeForColor( color );
		data[ x ][ y ].type = type;

		if( type == GRIDTYPE_UNKNOWN )
		{
			//VALIDATE_MSG5( type != GRIDTYPE_UNKNOWN, "Map data error: pixel (x,y) = (%d,%d), RGB color was %d,%d,%d; this is not a recognized color type!\n", x, y, color.r, color.g, color.b );
			type = GRIDTYPE_EMPTY;
		}

		/// Make this pixel floor-colored if it is of a non-visible type
		if( !gridDataTable[ type ].isVisible )
		{
			pixels[ startByteNumber + 0 ] = COLOR_EMPTY.b;
			pixels[ startByteNumber + 1 ] = COLOR_EMPTY.g;
			pixels[ startByteNumber + 2 ] = COLOR_EMPTY.r;
		}
	}
}



//-------------------------------------------------------------------------------------------------
// construction
//-------------------------------------------------------------------------------------------------
MapData::MapData( unsigned char* pixels, int width, int height )
{
	/// Parse image RGBs and convert to map metadata
	ParseImageMetaData( pixels, width, height );
}

//-------------------------------------------------------------------------------------------------
#define ROUND_DOUBLE(V) (floor((V + 0.5)))
bool MapData::TraceLine( int traceBoxRadius, const Vector2& start, const Vector2& end, int& intersectX, int& intersectY, GridDataType& intersectType, float& scanFraction )
{
	// find slope
	Vector2 delta = end - start;
	const float deltaLen = delta.Normalize();

	// trace line
	const int count = (int)deltaLen;
	Vector2 pos( start );
	intersectX = (int)end.x;
	intersectY = (int)end.y;
	scanFraction = 1.0f;

	for( int i = 0; i <= count; i++, pos += delta )
	{
		const int xPixel = (int)ROUND_DOUBLE( pos.x );
		const int yPixel = (int)ROUND_DOUBLE( pos.y );

		int minX = Max( MAP_MIN_X, xPixel - traceBoxRadius );
		int maxX = Min( MAP_MAX_X, xPixel + traceBoxRadius );
		int minY = Max( MAP_MIN_Y, yPixel - traceBoxRadius );
		int maxY = Min( MAP_MAX_Y, yPixel + traceBoxRadius );

		for( int scanX = minX; scanX <= maxX; scanX ++ )
		for( int scanY = minY; scanY <= maxY; scanY ++ )
		{
			intersectType = data[ scanX ][ scanY ].type;
			if( intersectType == GRIDTYPE_WALL || intersectType == GRIDTYPE_TEMPOCCLUDER )
			{
				intersectX = scanX;
				intersectY = scanY;
				pos -= start;
				scanFraction = pos.CalcLength() / deltaLen;
				return false;
			}
		}

	}
	return true;
}

//-------------------------------------------------------------------------------------------------
bool MapData::TraceLine( int traceBoxRadius, const Vector2& start, const Vector2& end )
{
	int a, b;
	GridDataType t;
	float f;
	return TraceLine( traceBoxRadius, start, end, a, b, t, f );
}