// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.

#pragma once
#include "Image.h"
#include "Vector2.h"

//-------------------------------------------------------------------------------------------------
// Grid contents types
//-------------------------------------------------------------------------------------------------
enum GridDataType
{
	GRIDTYPE_UNKNOWN = -1,
	GRIDTYPE_EMPTY = 0,
	GRIDTYPE_WALL,
	GRIDTYPE_ENEMYSTART,
	GRIDTYPE_ENEMYFACINGHINT,
	GRIDTYPE_ENEMYPATH,
	GRIDTYPE_PLAYERSTART,
	GRIDTYPE_PLAYERGOAL,
	GRIDTYPE_TEMPOCCLUDER,
	NUM_GRIDTYPES
};


struct Color
{
	unsigned char r;
	unsigned char g;
	unsigned char b;

	bool operator == ( const Color& rhs );
};


inline bool Color::operator ==( const Color& rhs )
{
	return( r == rhs.r &&
		g == rhs.g &&
		b == rhs.b );
}


#define COLOR_WHITE		{255,255,255}
#define COLOR_RED		{255,0,0}
#define COLOR_HALFRED	{128,0,0}
#define COLOR_BLACK		{0,0,0}
#define COLOR_GREEN		{0,255,0}
#define COLOR_YELLOW	{255,255,0}
#define COLOR_BLUE		{0,0,255}


const Color COLOR_EMPTY				= COLOR_BLACK;
const Color COLOR_WALL				= COLOR_WHITE;
const Color COLOR_ENEMYSTART		= COLOR_RED;
const Color COLOR_ENEMYFACINGHINT	= COLOR_HALFRED;
const Color COLOR_ENEMYPATH			= COLOR_BLUE;
const Color COLOR_PLAYERSTART		= COLOR_GREEN;
const Color COLOR_PLAYERGOAL		= COLOR_YELLOW;


struct GridDataAssociation
{
	GridDataType		type;
	Color				color;
	bool				isVisible; // false if this type & color represent meta-data only
	char*				name;
};



//-------------------------------------------------------------------------------------------------
// General enums
//-------------------------------------------------------------------------------------------------
enum
{
	MAP_WIDTH	= 1280,
	MAP_HEIGHT	= 960,

	MAP_MIN_X	= 0,
	MAP_MIN_Y	= 0,
	MAP_MAX_X	= MAP_WIDTH - 1,
	MAP_MAX_Y	= MAP_HEIGHT - 1,
};



//-------------------------------------------------------------------------------------------------
// Single grid info
//-------------------------------------------------------------------------------------------------
struct MapGridInfo
{
	GridDataType		type;
};



//-------------------------------------------------------------------------------------------------
// The entire map meta-data
//-------------------------------------------------------------------------------------------------
struct MapData
{
	//-------------------------------------------------------------------------------------------------
	// MEMBERS
	//-------------------------------------------------------------------------------------------------
	MapGridInfo		data[ MAP_WIDTH ][ MAP_HEIGHT ];

	//-------------------------------------------------------------------------------------------------
	// METHODS
	//-------------------------------------------------------------------------------------------------
	MapData( unsigned char* pixels, int width, int height );
	ParseImageMetaData( unsigned char* pixels, int width, int height );
	GridDataType GetMapDataTypeForColor( Color& color );

	// traceline returns false if the trace did not complete without hitting a wall
	bool TraceLine( int traceBoxRadius, const Vector2& start, const Vector2& end );
	bool TraceLine( int traceBoxRadius, const Vector2& start, const Vector2& end, int& intersectX, int& intersectY, GridDataType& intersectType, float& scanFraction );
};



//-------------------------------------------------------------------------------------------------
// exported global data
//-------------------------------------------------------------------------------------------------
extern MapData* g_mapData;


