// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.

#include "pch.h"
#include "MapData.h"
#include "PathFinding.h"


//-------------------------------------------------------------------------------------------------
// global data
//-------------------------------------------------------------------------------------------------
PathFinding g_pathFinding;

const float g_semiCardinalCos[ 8 ] =
{
	1.0f,
	0.70710678f,
	0.0f,
	-0.70710678f,
	-1.0f,
	-0.70710678f,
	0.0f,
	0.70710678f,
};

const float g_semiCardinalSin[ 8 ] =
{
	0.0f,
	-0.70710678f,
	-1.0f,
	-0.70710678f,
	0.0f,
	0.70710678f,
	1.0f,
	0.70710678f,
};

const int g_semiCardinalX[ 8 ] =
{
	1, 1, 0, -1, -1, -1, 0, 1,
};

const int g_semiCardinalY[ 8 ] =
{
	0, -1, -1, -1, 0, 1, 1, 1, 
};

const float VALUE0 = 0.0f;
const float VALUE1 = 0.44721360f;
const float VALUE2 = 0.70710678f;
const float VALUE3 = 0.89442719f;
const float VALUE4 = 1.0f;


const float g_quarterCardinalCos[ 16 ] =
{
	VALUE4, // 0 DEGREES
	VALUE3,
	VALUE2,
	VALUE1,
	VALUE0, // 90 DEGREES (BACKWARDS)
	-VALUE1,
	-VALUE2,
	-VALUE3,
	-VALUE4, // 180
	-VALUE3,
	-VALUE2,
	-VALUE1,
	VALUE0, // 270
	VALUE1,
	VALUE2,
	VALUE3,
};

const float g_quarterCardinalSin[ 16 ] =
{
	VALUE0, // 0 DEGREES (BACKWARDS)
	-VALUE1,
	-VALUE2,
	-VALUE3,
	-VALUE4, // 90
	-VALUE3,
	-VALUE2,
	-VALUE1,
	VALUE0, // 180
	VALUE1,
	VALUE2,
	VALUE3,
	VALUE4, // 270
	VALUE3,
	VALUE2,
	VALUE1,
};
	
/*
0.44721359549995793928183473374626
0.89442719099991587856366946749251
*/

const int g_quarterCardinalX[ 16 ] =
{
	2, 2, 2, 1, 0, -1, -2, -2, -2, -2, -2, -1, 0, 1, 2, 2,
};


const int g_quarterCardinalY[ 16 ] =
{
	0, -1, -2, -2, -2, -2, -2, -1, 0, 1, 2, 2, 2, 2, 2, 1,
};







//-------------------------------------------------------------------------------------------------
// IsXYOnEnemyPath
//-------------------------------------------------------------------------------------------------
bool PathFinding::IsXYOnEnemyPath( int x, int y )
{
	return( g_mapData->data[ x ][ y ].type == GRIDTYPE_ENEMYPATH );
}


//-------------------------------------------------------------------------------------------------
// FindClosestPathLocation
//-------------------------------------------------------------------------------------------------
void PathFinding::FindClosestPathLocation( int& closestPathX, int& closestPathY, const Vector2& referencePosition )
{
	int posX = (int) referencePosition.x;
	int posY = (int) referencePosition.y;

	/// Early out for already-on-the-path case
	if( IsXYOnEnemyPath( posX, posY ) )
	{
		closestPathX = posX;
		closestPathY = posY;
		return;
	}

	/// Search in concentric rings (er, squares), going out from the origin point
	for( int searchRadius = 1; searchRadius < MAX_PATHFINDING_SEARCH_RADIUS; searchRadius ++ )
	{
		int minX = Max( MAP_MIN_X, posX - searchRadius );
		int maxX = Min( MAP_MAX_X, posX + searchRadius );
		int minY = Max( MAP_MIN_Y, posY - searchRadius );
		int maxY = Min( MAP_MAX_Y, posY + searchRadius );

		/// Search in a ring (er, square) at this radius
		for( int x = minX; x <= maxX; x ++ )
		{
			if( g_mapData->data[ x ][ minY ].type == GRIDTYPE_ENEMYPATH )
			{
				closestPathX = x;
				closestPathY = minY;
				return;
			}

			if( g_mapData->data[ x ][ maxY ].type == GRIDTYPE_ENEMYPATH )
			{
				closestPathX = x;
				closestPathY = maxY;
				return;
			}
		}

		for( int y = minY; y <= maxY; y ++ )
		{
			if( g_mapData->data[ minX ][ y ].type == GRIDTYPE_ENEMYPATH )
			{
				closestPathX = minX;
				closestPathY = y;
				return;
			}

			if( g_mapData->data[ maxX ][ y ].type == GRIDTYPE_ENEMYPATH )
			{
				closestPathX = maxX;
				closestPathY = y;
				return;
			}
		}
	}

	/// Didn't find a closest path; leave it where it is and raise a ruckus
	char msg[ 1024 ];
	sprintf( msg, "Enemy at %d,%d (rgb=%d,%d,%d) couldn't find a nearby path (rgb=%d,%d,%d) to jump to within a search radius of %d pixels\n",
		posX, posY,
		COLOR_ENEMYSTART.r, COLOR_ENEMYSTART.g, COLOR_ENEMYSTART.b,
		COLOR_ENEMYPATH.r, COLOR_ENEMYPATH.g, COLOR_ENEMYPATH.b,
		MAX_PATHFINDING_SEARCH_RADIUS );
	PANIC_MSG0( msg );
	closestPathX = posX;
	closestPathY = posY;

}


